import json
import etcd

class EtcdDocument(object):

    def __init__(self, client, root_path, root=None):
        if client is not None and root is None:
            try:
                self.root = client.read(root_path, recursive=True)
            except:
                self.root = None
        else:
            self.root = root

    def _get_node(self, path):
        if self.root is None:
            return None
        d = self.root
        for s in path:
            if s == ".":
                pass
            else:
                if hasattr(d, "_children"):
                    children = d._children
                    for child in children:
                        if child["key"].split("/")[-1] == s:
                            d = etcd.EtcdResult(None, node=child)
                            break
                    else:
                        return None
                else:
                    return None
        return d

    def get_value(self, path):
        node = self._get_node(path)
        try:
            val = node.value
        except AttributeError:
            val = None
        if val is not None:
            try:
                return json.loads(val)
            except:
                return val
        else:
            return val 

    def subdocument(self, path):
        return EtcdDocument(None, None, root=self._get_node(path))

    def children(self):
        if self.root is None or not hasattr(self.root, "_children"):
            return []
        else:
            return [(child["key"].split("/")[-1], EtcdDocument(None, None, root=etcd.EtcdResult(None, node=child))) for child in self.root._children]



#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import dbus
import module
import logging
import cachetools
import psutil

class PropWrapper(object):

    def __init__(self, interface, props):
        self.props = props
        self.interface = interface

    def __getattr__(self, name):
        return self.props.Get(self.interface, name)

@cachetools.cached(cache=cachetools.LRUCache(5000))
def _get_process(pid):    
    return psutil.Process(pid=pid)

class CGroupWrapper(object):

    """
    * {{...cpu_time_pct}} - percent share of CPU core time by the given cgroup
    * {{...rss}} - total RSS memory in bytes by the given cgroup 
    """

    def __init__(self, cgname):
        self.cgname = cgname

    def _get_procs(self):
        result = []
        with open("/sys/fs/cgroup/systemd%s/cgroup.procs" % self.cgname) as fp:
            for pid in fp:
                result.append(int(pid))
        return result

    @property
    def cpu_time_pct(self):
        procs = self._get_procs()
        cpu_percent = 0.0
        for pid in procs:
            # The line below will get the CPU usage pct
            # since the last invocation of cpu_percent(). 
            # The important point here is memoization of the
            # process object. If the process object is brand
            # new, then cpu_percent() returns 0. The memoization
            # occurs in _get_process() and is mostly controlled 
            # by the LRUCache settings. If the number of monitored
            # processes approaches 5000 (see decorator), then
            # the returned value becomes incorrect
            cpu_percent += _get_process(pid).cpu_percent()
        return cpu_percent

    @property
    def rss(self):
        procs = self._get_procs()
        memory = 0.0
        for pid in procs:
            memory += _get_process(pid).memory_info().rss
        return memory

class UnitWrapper(object):

    """
    * {{...properties}} - return the object which contains unit properties 
      (see https://wiki.freedesktop.org/www/Software/systemd/dbus/)
    * {{...cgroup}} - return the CGroup properties representation (see CGroupWrapper)
    * {{...handle(action)}} - execute unit action (one of start, stop, restart)
    """

    def __init__(self, obj):
        self.unit = dbus.Interface(obj, dbus_interface="org.freedesktop.systemd1.Unit")
        self.props = PropWrapper(
            "org.freedesktop.systemd1.Unit", 
            dbus.Interface(obj, dbus_interface="org.freedesktop.DBus.Properties"))
        self.service_props = PropWrapper(
            "org.freedesktop.systemd1.Service",
            dbus.Interface(obj, dbus_interface="org.freedesktop.DBus.Properties"))
        self.cgroup = CGroupWrapper(str(self.service_props.ControlGroup))

    def __getattr__(self, name):
        if name == "properties":
            return self.props
        elif name == "cgroup":
            return self.cgroup
        elif name == "handle":
            def handle(action):
                if action == "start":
                    self.unit.Start("replace")
                elif action == "stop":
                    self.unit.Stop("replace")
                elif action == "restart":
                    self.unit.Restart("replace")
                else:
                    raise ValueError(action)                    
            return handle
        else:        
            return getattr(self.unit, name)


class Systemd(module.BaseModule):

    """
    * {{unit(name)}} - get the systemd unit object (see UnitWrapper)
    """    

    SYSTEMD = "org.freedesktop.systemd1"

    def __init__(self):
        self.bus = dbus.SystemBus()
        systemd1 = self.bus.get_object(Systemd.SYSTEMD, "/org/freedesktop/systemd1")
        self.manager = dbus.Interface(systemd1, dbus_interface="org.freedesktop.systemd1.Manager")
        self.manager_props =  PropWrapper("org.freedesktop.systemd1.Manager", 
            dbus.Interface(systemd1, dbus_interface="org.freedesktop.DBus.Properties"))

    def keys(self):
        return ["unit" ,"reboot", "boot_time"]

    def get(self, key):
        if key == "unit":
            def unit(name):
                unit_path = self.manager.GetUnit(name)
                unit_obj = self.bus.get_object(Systemd.SYSTEMD, unit_path)
                return UnitWrapper(unit_obj)
            return unit
        elif key == "reboot":
            def reboot():
                self.manager.Reboot()
            return reboot
        elif key == "boot_time":
            return self.manager_props.KernelTimestamp/1000000.0
        else:
            raise KeyError(key)



#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import dbus
import module
import logging

class OS(module.BaseModule):

    """
    * {{os_name}} - get the pretty printed OS name
    """

    HOSTNAME = "org.freedesktop.hostname1"

    def __init__(self):
        self.bus = dbus.SystemBus()
        hostname1 = self.bus.get_object(OS.HOSTNAME, "/org/freedesktop/hostname1")
        self.hostname_props = dbus.Interface(hostname1, dbus_interface="org.freedesktop.DBus.Properties")        

    def keys(self):
        return ["os_name"]

    def get(self, key):
        if key == "os_name":
            return self.hostname_props.Get("org.freedesktop.hostname1", "OperatingSystemPrettyName")
        else:
            raise KeyError(key)



#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import psutil
import time
import module
import socket

class Host(module.BaseModule):

    """
    * {{addr(family)}} - get address for the given address family 
      (one of ip, ip6, link)
    * {{ip}} - get IPv4 address 
    """

    def __init__(self, ttl=300):        
        self.interfaces = None
        self.ttl = ttl

    def keys(self):
        return ["addr", "ip"]

    def get(self, key):        
        def addr(device, address_family):
            if address_family == "ip":
                af = socket.AF_INET
            elif address_family == "ip6":
                af = socket.AF_INET6
            elif address_family == "link":
                af = psutil.AF_LINK
            else:
                raise ValueError("Unsupported address family: %s" % address_family)
            
            curtime = time.time()
            if self.interfaces is None or curtime - self.last_query > self.ttl:
                self.interfaces = psutil.net_if_addrs()
                self.last_query = curtime

            for addr in self.interfaces[device]:
                if addr.family == af:
                    return addr.address

            raise ValueError("Device %s doesn't have address of the address family %s" % (device, address_family))
        def ip(device):
            return addr(device, "ip")
        if key == "addr":
            return addr
        if key == "ip":                
            return ip
        else:
            raise KeyError()


#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import psutil
import module

class Memory(module.BaseModule):

    """
    * {{memory_usage_pct}} - physical memory usage (percent)
    * {{total_memory}} - total phycal memory
    * {{available_memory}} - available memory
    """

    def keys(self):
        return ["memory_usage_pct", "total_memory", "available_memory"]

    def get(self, key):
        if key == "memory_usage_pct":
            mem = psutil.virtual_memory()
            return 1 - float(mem.available) / mem.total
        elif key == "total_memory":
            mem = psutil.virtual_memory()
            return float(mem.total)
        elif key == "available_memory":
            mem = psutil.virtual_memory()
            return float(mem.available)
        else:
            raise KeyError()


#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import logging
import contextlib
import UserDict
import threading

class Client(object):

    def write(self, name, value, ttl=None):
        assert False

    def close(self):
        pass

    def add_handler(self, context, defaults, name, ntemp, vtemp):
        pass

class StdoutClient(Client):

    def write(self, name, value, ttl=None):
        print "%s = %s" % (name, value)

class EtcdClient(Client):

    def __init__(self, host, port):
        import etcd
        self.client = etcd.Client(host=host, port=port)

    def write(self, name, value, ttl=None):
        self.client.write(name, value, ttl=ttl)

    def close(self):
        pass

    def add_handler(self, context, defaults, name, ntemp, vtemp):
        EtcdHandlerThread(name, ntemp, vtemp, self.client, defaults, context).start()

class EtcdHandlerThread(threading.Thread):

    def __init__(self, name, ntemp, vtemp, client, defaults, context):
        super(EtcdHandlerThread, self).__init__(name=name)
        self.daemon = True
        self.ntemp = ntemp
        self.vtemp = vtemp
        self.client = client
        self.defaults = defaults
        self.context = context

    def run(self):
        key = self.ntemp.render(self.context)
        defaults = self.defaults.render(self.context)
        for event in self.client.eternal_watch(key=key,recursive=True):
            if event.action == "create":
                ctx = dict(self.context)
                ctx.update(defaults)
                ctx["event"] = event.value
                try:
                    self.vtemp.render(ctx)
                except:
                    logging.error("Error processing action on %s" % key, exc_info=True)


class Context(UserDict.DictMixin):

    def __init__(self):
        self.modules = []

    def __getitem__(self, key):
        for module in self.modules:
            if module.provides(key):
                return module.get(key)
        raise KeyError(key)

    def keys(self):
        return [key for module in self.modules for key in module.keys()]

    def add_module(self, module):
        self.modules.append(module)   

class Template(object):

    def __init__(self, repr, template):
        self.repr = repr
        self.template = template

    def __repr__(self):
        return self.repr 

    def render(self, context):
        return self.template.render(context)            

class EtcdStat(object):

    def __init__(self, url, interval, items, handlers, defaults):
        from urlparse import urlparse
        from jinja2 import Environment
        env = Environment()
        self.items = { n : (Template(n, env.from_string(n)), Template(v, env.from_string(v))) for (n,v) in items }
        self.handlers = { n : (Template(n, env.from_string(n)), Template(v, env.from_string(v))) for (n,v) in handlers }
        self.interval = interval

        parsed_url = urlparse(url)
        if parsed_url.scheme == "stdout":
            self.client = StdoutClient()
        elif parsed_url.scheme == "http":
            self.client = EtcdClient(host=parsed_url.hostname, port=parsed_url.port if parsed_url.port else 80)
        else:
            raise ValueError("Unsupported scheme" % parsed_url.scheme)

        self.defaults = TemplateDict({ n : Template(v, env.from_string(v)) for (n,v) in defaults })

        self.context = Context()

        import cpu
        import memory
        import disk
        import host
        import systemd
        import os_support
        self.context.add_module(cpu.Cpu())
        self.context.add_module(memory.Memory())
        self.context.add_module(disk.Disk())                
        self.context.add_module(host.Host())
        self.context.add_module(systemd.Systemd())
        self.context.add_module(os_support.OS())

        for n in self.handlers:
            self.client.add_handler(self.context, self.defaults, n, *(self.handlers[n]))

    def update_etcd(self):
        import jinja2

        defaults = self.defaults.render(self.context)
        #print "defaults =", defaults 
        ctx = dict(self.context)
        ctx.update(defaults)

        for p in self.items:
            (name, value) = self.items[p]
            try:
                _name = name.render(ctx)
            except:
                logging.error("Error rendering %s" % name, exc_info=True)
                continue
            try:
                _value = value.render(ctx)
            except:
                logging.error("Error rendering %s" % value, exc_info=True)
                continue
            try:
                self.client.write(_name, _value, ttl=int(self.interval*2))
            except:
                logging.error("Error writing %s" % _name, exc_info=True)
                continue

    def run(self):
        import time        
        while True:
            try:
                self.update_etcd()
            except:
                logging.error("Error", exc_info=True)
                pass
            time.sleep(self.interval)

    def close(self):
        self.client.close()

class TemplateDict(object):

    def __init__(self, src):
        self.items = src

    def render(self, context):
        return { name: value.render(context) for (name, value) in self.items.iteritems() }

def main():
    import argparse
    import complexini
    import os.path

    parser = argparse.ArgumentParser()
    parser.add_argument("url", metavar="URL", help="Etcd URL", default="http://localhost:2379")
    parser.add_argument("-c", "--config", metavar="CONFIG", help="Configuration file", default="/etc/etcdstat.cfg")
    parser.add_argument("-i", "--interval", metavar="INTERVAL", help="Poll interval (sec)", type=float, default="10")

    args = parser.parse_args()
    
    config_dir = os.path.dirname(args.config)
    config = complexini.ComplexIniFile(root_dir=config_dir if config_dir else None)
    config.read(args.config)

    defaults = []
    for sect in ["Defaults"]:
        for name, value in config.items(sect):
            defaults.append((name, value))
    #print defaults

    items = []
    for sect in ["System","Services"]:
        for name, value in config.items(sect):
            items.append((name, value))

    handlers = []
    for sect in ["Handlers"]:
        for name, value in config.items(sect):
            handlers.append((name, value))
    
    with create_etcdstat(args.url, args.interval, items, handlers, defaults) as etcdstat:
        etcdstat.run()

@contextlib.contextmanager
def create_etcdstat(url, interval, items, handlers, defaults):
    etcdstat = EtcdStat(url, interval, items, handlers, defaults)
    try:
        yield etcdstat
    finally:
        etcdstat.close()

if __name__ == "__main__":
    main()        


import os.path
import ConfigParser

class ComplexIniFile(object):

    def __init__(self, root_dir=None):
        self.root_dir = root_dir
        self.root = None
        self.parsers = []

    def read(self, file):
        self.root = ConfigParser.ConfigParser()
        self.root.optionxform = str # preserve case
        self.root.read(file)

        try:
            for name, value in self.root.items("Includes"):
                if name != "include":
                    raise ValueError("Invalid entry in `Includes` section: %s: %s" % (name, value))
                for file in value.split(","):
                    path = os.path.join(self.root_dir, file) if self.root_dir else file
                    sub_parser = ComplexIniFile(root_dir=self.root_dir)
                    sub_parser.read(path)
                    self.parsers.append(sub_parser)
        except ConfigParser.NoSectionError:
            pass

    def items(self, section):
        processed_names = set()
        try:
            for name, value in self.root.items(section):
                processed_names.add(name)
                yield (name, value)
        except ConfigParser.NoSectionError:
            pass
        for p in self.parsers:
            for name, value in p.items(section):
                if name not in processed_names:
                    processed_names.add(name)
                    yield (name, value)

if __name__ == "__main__":

    c = ComplexIniFile()
    c.read("test.cfg")
    for n, v in c.items("System"):
        print n, v
    



#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import psutil
import module
import time

class Cpu(module.BaseModule):

    def __init__(self):
        self.cpu = None        

    def keys(self):
        return ["cpu"]

    def get(self, key):
        if key == "cpu":
            curtime = time.time()
            if self.cpu is None or curtime - self.last_query > 10:
                self.cpu = psutil.cpu_percent(interval=None)
                self.last_query = curtime 
            return float(self.cpu) / 100
        else:
            raise KeyError()


#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import psutil
import module

class Disk(module.BaseModule):

    def keys(self):
        return ["disk_usage_pct", "available_storage", "total_storage"]

    def get(self, key):
        if key == "disk_usage_pct":
            def disk_usage(path):
                return float(psutil.disk_usage(path).percent)/100
            return disk_usage
        elif key == "available_storage":
            def available_storage(path):
                return psutil.disk_usage(path).free
            return available_storage    
        elif key == "total_storage":
            def total_storage(path):
                return psutil.disk_usage(path).total
            return total_storage    
        else:
            raise KeyError()


# -*- coding: utf-8 -*-

#    Copyright 2017 Thingswise, LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

 
 
import re
from setuptools import setup
import versioneer

with open("README.rst", "rb") as f:
    long_descr = f.read().decode("utf-8")
  
setup(
    cmdclass=versioneer.get_cmdclass(),
    name = "tw-etcdstat",
    packages = ["etcdstat", "etcdstat.etcdparser"],
    entry_points = {
        "console_scripts": ['etcdstat = etcdstat.etcdstat:main']
    },
    version=versioneer.get_version(),
    description = "Python system and service monitoring service",
    long_description = long_descr,
    author = "Alexander Lukichev",
    author_email = "alexander.lukichev@thingswise.com",
    url = "https://github.com/thingswise/tw-etcdstat",
    install_requires = [
        "jinja2",
        "psutil",
        "cachetools",
        "python-etcd"
    ],
    license="Apache2"
)



class TreeParser(object):

    def __init__(self, tmpl):
        self.root = parse_tree(tmpl)

    def render(self, input_doc):
        return self.root.render(Context(input_doc, {}))

def parse_tree(tmpl):
    if isinstance(tmpl, dict):
        result = ObjectValueTemplate()
        for n, v in tmpl.iteritems():
            if len(n) > 3 and n[0] == "^" and n[1] == "(":
                i = n.find(")", 2)
                if i < 0:
                    raise ValueError("Invalid template: %s" % n)
                path = n[2:i].strip()
                name = n[i+1:].strip()
                if len(name) <= 0:
                    raise ValueError("Invalid template: %s" % n)
                result.put(NameTemplate(path_template=PathTemplate(path), name_template=StrTemplate(name)), parse_tree(v))
            else:
                result.put(NameTemplate(name_template=StrTemplate(n)), parse_tree(v))
        return result
    elif isinstance(tmpl, list):
        result = ArrayValueTemplate()
        for v in tmpl:
            result.append(parse_tree(v))
        return result
    elif isinstance(tmpl, str):
        if len(tmpl) > 3 and tmpl[0] == "^" and tmpl[1] == "(" and tmpl[-1] == ")":
            path = tmpl[2:-1].strip()
            if len(path) <= 0:
                raise ValueError("Invalid template: %s" % tmpl)
            return ReferenceValueTemplate(PathTemplate(path))
        else:
            return PrimitiveValueTemplate(tmpl)
    else:
        return PrimitiveValueTemplate(tmpl)

class ValueTemplate(object):
    pass

class ObjectValueTemplate(ValueTemplate):

    def __init__(self):
        self.values = []

    def render(self, context):
        result = {}
        for n, v in self.values:
            for name, subctx in n.render(context):
                val = v.render(subctx)
                if val is None:
                    if name in result:
                        del result[name]
                else:
                    result[name] = v.render(subctx)
        return result

    def put(self, n, v):
        self.values.append((n, v))

class ArrayValueTemplate(ValueTemplate):

    def __init__(self):
        self.values = []

    def render(self, context):
        return [ v.render(context) for v in self.values ]

    def append(self, v):
        self.values.append(v)

class PrimitiveValueTemplate(ValueTemplate):

    def __init__(self, value):
        self.value = value

    def render(self, context):        
        return self.value

class ReferenceValueTemplate(ValueTemplate):

    def __init__(self, path_template):
        self.path = path_template

    def render(self, context):
        return context.get_value(self.path)

class NameTemplate(object):

    def __init__(self, name_template, path_template=None):                        
        self.path_template = path_template
        self.name_template = name_template

    def render(self, context):
        if self.path_template is None:
            yield context.render_string(self.name_template), context
        else:
            for subctx in context.match(self.path_template):
                name = subctx.render_string(self.name_template)
                yield name, subctx

class Placeholder(object):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "{%s}" % self.name

    def render(self, context):
        return context.get_var(self.name)

def parse_path(path):
    p = path.split("/")
    if len(p) == 0:
        return p
    if p[0] == "":
        p = p[1:]
    if len(p) == 0:
        return p
    if p[-1] == "":
        p = p[:-1]
    if len(p) == 0:
        return p
    return [ Placeholder(i[1:-1]) if len(i) >= 2 and i[0] == "{" and i[-1] == "}" else i for i in p ]                

class PathTemplate(object):

    def __init__(self, path):
        self.items = parse_path(path)

    def render(self, context):
        return [s.render(context) if type(s) == Placeholder else s for s in self.items]

def parse_str(d):
    result = []
    i = d.find("{")
    s = 0
    while i != -1:
      result.append(d[s:i])
      j = d.find("}", i+1)
      if j == -1:
        raise ValueError("Invalid key def: %s" % d)
      else:
        n = d[i+1:j]
        result.append(Placeholder(n))
        s = j+1
        i = d.find("{", s)
    result.append(d[s:])
    return result

class StrTemplate(object):

    def __init__(self, template):
        self.template = parse_str(template)

    def render(self, context):
        result = ""
        for s in self.template:
            if type(s) == str:
                result += s
            elif type(s) == Placeholder:
                result += s.render(context)
            else:
                assert False
        return result

def _match(path_template_arr, root, variables):
    #print "_match: %s" % (path_template_arr)
    i = 0
    while i < len(path_template_arr) and type(path_template_arr[i]) == str:
        i += 1
    if i > 0:
        root = root.subdocument(path_template_arr[:i])
    if i >= len(path_template_arr):
        #print "==> %s" % root
        yield Context(root, variables)
    else:
        for name, child in root.children():
            subctx = variables.copy()
            assert type(path_template_arr[i]) == Placeholder
            subctx[path_template_arr[i].name] = name
            for ctx in _match(path_template_arr[i+1:], child, subctx):
                #print "==> %s" % ctx.root
                yield ctx

class Context(object):

    def __init__(self, root, variables):
        self.root = root        
        self.variables = variables

    def get_var(self, name):
        return self.variables[name]

    def render_string(self, name_template):
        return name_template.render(self)

    def get_value(self, path_template):
        return self.root.get_value(path_template.render(self))

    def match(self, path_template):
        for ctx in _match(path_template.items, self.root, self.variables):
            yield ctx

class DictDocument(object):

    def __init__(self, d):
        self.val = d

    def get_value(self, path):
        d = self.val
        for s in path:
            if s == ".":
                pass
            else:
                try:
                    d = d[s]
                except KeyError:
                    return None
        return d

    def subdocument(self, path):
        return DictDocument(self.get_value(path))

    def children(self):
        if self.val is None or not isinstance(self.val, dict):
            return []
        else:
            return [(n, DictDocument(v)) for n, v in self.val.iteritems()]

    def __repr__(self):
        return json.dumps(self.val, indent=2)

if __name__ == "__main__":

    import json
    import sys
    import yaml

    with open(sys.argv[1]) as fp:
        tmpl = yaml.load(fp)

    with open(sys.argv[2]) as fp:
        input_doc = json.load(fp)        

    result = TreeParser(tmpl).render(DictDocument(input_doc))
    print result
    print json.dumps(result, indent=2)

    